import sys,re

newline=re.compile(r"[\r\n]+")

def nonempty_partials(x):
  if len(x)<2:
    return []
  else:
    return [x[:1]+y for y in nonempty_partials(x[1:])]+ \
           [y for y in nonempty_partials(x[1:])]+ \
           [x[:1],x[1:]]

class mnmx:
  def __init__(self):
    self.mn=None
    self.mx=None

class status(mnmx):
  def __init__(self,mn,mx,ancestor,sibling,descendant):
    self.mn=mn[:]
    self.mx=mx[:]
    self.changed_a=[]
    self.changed_s=[]
    self.changed_d=[]
    self.ancestor=ancestor[:]
    self.sibling=sibling[:]
    self.descendant=descendant[:]
    self.explain=[]

class constraint_topology(mnmx):
  def __init__(self):
    self.ancestor=[]
    self.sibling=[]
    self.descendant=[]
    self.size=0
    self.mn=None
    self.mx=None

class topology:
  def __init__(self,list_to_num):
    self.constraints={}
    self.atoms=0
    self.maxsize=0
    self.list_to_num=list_to_num
  def add_constraint(self,list,mn,mx):
    num=self.list_to_num(list)
    if num not in self.constraints:
      self.constraints[num]=constraint_topology()
    if self.constraints[num].size!=0:
      return # constraint already added
    self.constraints[num].size=len(list)
    if len(list)>self.maxsize:
      self.maxsize=len(list)
    if len(list)==1:
      self.atoms+=1
    self.constraints[num].mn=mn[:]
    self.constraints[num].mx=mx[:]
    for i in nonempty_partials(list):
      descendant=self.list_to_num(i)
      sibling=num-descendant
      self.constraints[num].descendant.append((descendant,sibling))
      if descendant not in self.constraints:
        self.constraints[descendant]=constraint_topology()
      self.constraints[descendant].sibling.append((sibling,num))
      if mn!=mx:
        self.constraints[descendant].ancestor.append((num,sibling))

class queue:
  def __init__(self,t):
    self.events=[[] for i in range(t.maxsize+1)]
    self.minsize=1
    self.topology=t
  def enqueue(self,event):
    size=self.topology.constraints[event].size
    if event not in self.events[size]:
      self.events[size].append(event)
    if self.minsize>size:
      self.minsize=size
  def dequeue(self):
    while self.minsize<len(self.events) and len(self.events[self.minsize])==0:
      self.minsize+=1
    if self.minsize==len(self.events):
      return None
    else:
      rc=self.events[self.minsize][0]
      self.events[self.minsize]=self.events[self.minsize][1:]
      return rc

class solver:
  def __init__(self,t,output,verbose):
    self.topology=t
    self.constraints={}
    for c in self.topology.constraints:
      self.constraints[c]= \
        status(self.topology.constraints[c].mn,self.topology.constraints[c].mx,\
               self.topology.constraints[c].ancestor, \
               self.topology.constraints[c].sibling, \
               self.topology.constraints[c].descendant)
    self.oracle=[]
    self.pending=queue(t)
    self.unsolved_atoms=t.atoms
    self.output=output
    self.verbose=verbose
  def update(self,event,mn,mx):
    true_update=False
    for i in range(len(mn)):
      if mn[i]>self.constraints[event].mn[i]:
        true_update=True
        self.constraints[event].mn[i]=mn[i]
      if mx[i]<self.constraints[event].mx[i]:
        true_update=True
        self.constraints[event].mx[i]=mx[i]
      if self.constraints[event].mn[i]>self.constraints[event].mx[i]:
        raise "Inconsistent"
    if true_update:
      smn=sum(self.constraints[event].mn)
      smx=sum(self.constraints[event].mx)
      if smn>self.topology.constraints[event].size or \
         smx<self.topology.constraints[event].size:
        raise "Inconsistent"
      if smn==self.topology.constraints[event].size:
        for i in range(len(self.constraints[event].mx)):
          if self.constraints[event].mx[i]>self.constraints[event].mn[i]:
            self.constraints[event].mx[i]=self.constraints[event].mn[i]
            self.constraints[event].explain.append((0,i,0,event))
      elif smx==self.topology.constraints[event].size:
        for i in range(len(self.constraints[event].mn)):
          if self.constraints[event].mn[i]<self.constraints[event].mx[i]:
            self.constraints[event].mn[i]=self.constraints[event].mx[i]
            self.constraints[event].explain.append((1,i,0,event))
      if sum(self.constraints[event].mx)==1:
        self.unsolved_atoms-=1
        if self.verbose:
          self.output()
      for (ancestor,sibling) in self.constraints[event].ancestor:
        if event not in self.constraints[ancestor].changed_d:
          self.constraints[ancestor].changed_d.append((event,sibling))
        self.pending.enqueue(ancestor)
      for (sibling,ancestor) in self.constraints[event].sibling:
        if event not in self.constraints[sibling].changed_s:
          self.constraints[sibling].changed_s.append((event,ancestor))
        self.pending.enqueue(sibling)
      for (descendant,sibling) in self.constraints[event].descendant:
        if event not in self.constraints[descendant].changed_a:
          self.constraints[descendant].changed_a.append((event,sibling))
        self.pending.enqueue(descendant)
      if self.constraints[event].mn==self.constraints[event].mx:
        for (ancestor,sibling) in self.constraints[event].ancestor:
          self.constraints[ancestor].descendant.remove((event,sibling))
        for (sibling,ancestor) in self.constraints[event].sibling:
          self.constraints[sibling].sibling.remove((event,ancestor))
        for (descendant,sibling) in self.constraints[event].descendant:
          self.constraints[descendant].ancestor.remove((event,sibling))
    return true_update
  def dry_run(self,event):
    mn=self.constraints[event].mn[:]
    mx=self.constraints[event].mx[:]
    for (a,s) in self.constraints[event].changed_a:
      for i in range(len(mn)):
        if mn[i]<self.constraints[a].mn[i]-self.constraints[s].mx[i]:
          mn[i]=self.constraints[a].mn[i]-self.constraints[s].mx[i]
          self.constraints[event].explain.append((1,i,1,a))
        if mx[i]>self.constraints[a].mx[i]-self.constraints[s].mn[i]:
          mx[i]=self.constraints[a].mx[i]-self.constraints[s].mn[i]
          self.constraints[event].explain.append((0,i,1,a))
    for (s,a) in self.constraints[event].changed_s:
      for i in range(len(mn)):
        if mn[i]<self.constraints[a].mn[i]-self.constraints[s].mx[i]:
          mn[i]=self.constraints[a].mn[i]-self.constraints[s].mx[i]
          self.constraints[event].explain.append((1,i,2,s))
        if mx[i]>self.constraints[a].mx[i]-self.constraints[s].mn[i]:
          mx[i]=self.constraints[a].mx[i]-self.constraints[s].mn[i]
          self.constraints[event].explain.append((0,i,2,s))
    for (d,s) in self.constraints[event].changed_d:
      for i in range(len(mn)):
        if mn[i]<self.constraints[d].mn[i]+self.constraints[s].mn[i]:
          mn[i]=self.constraints[d].mn[i]+self.constraints[s].mn[i]
          self.constraints[event].explain.append((1,i,3,d))
        if mx[i]>self.constraints[d].mx[i]+self.constraints[s].mx[i]:
          mx[i]=self.constraints[d].mx[i]+self.constraints[s].mx[i]
          self.constraints[event].explain.append((0,i,3,d))
    self.constraints[event].changed_a=[]
    self.constraints[event].changed_s=[]
    self.constraints[event].changed_d=[]
    return (mn,mx)
  def solve(self):
    changed=False
    while len(self.oracle)!=0:
      for i in range(len(self.oracle[0][1])):
        self.constraints[self.oracle[0][0]].explain.append(
          (self.oracle[0][1][i],i,4,self.oracle[0][0]))
      if self.update(self.oracle[0][0],self.oracle[0][1],self.oracle[0][2]):
        changed=True
      self.oracle=self.oracle[1:]
      event=self.pending.dequeue()
      while self.unsolved_atoms>0 and event!=None:
        (mn,mx)=self.dry_run(event)
        if self.update(event,mn,mx):
          changed=True
        event=self.pending.dequeue()
    return changed

class sudoku:
  def __init__(self,n=3):
    self.n=n
    self.numbers=self.n*self.n
    self.rows=self.n*self.n
    self.columns=self.n*self.n
    self.area=self.rows*self.columns
    self.sectors=self.n*self.n
    basic_constraints=[]
    for i in range(self.numbers):
      basic_constraints.append([(x,i) for x in range(self.numbers)])
      basic_constraints.append([(i,x) for x in range(self.numbers)])
    for i in range(self.n):
      for j in range(self.n):
        basic_constraints.append([(self.n*i+x,self.n*j+y) \
          for x in range(self.n) for y in range(self.n)])
    self.sudoku_topology=topology(self.list_to_num)
    for bc in basic_constraints:
      self.sudoku_topology.add_constraint(bc,[1]*self.numbers,[1]*self.numbers)
      for s in nonempty_partials(bc):
        self.sudoku_topology.add_constraint(s,[0]*self.numbers,[1]*self.numbers)
    if n==3:
      self.dictionary={"1":0,"2":1,"3":2,"4":3,"5":4,"6":5,"7":6,"8":7,"9":8}
      self.inv_dict=["1","2","3","4","5","6","7","8","9"]
    else:
      self.dictionary={}
      self.inv_dict=[]
  def pos_to_num(self,x,y):
    return 2**(x+self.columns*y)
  def list_to_num(self,x):
    return sum([self.pos_to_num(y[0],y[1]) for y in x])
  def num_to_list(self,x):
    rc=[]
    i=0
    y=x
    while y>0:
      if y%2==1:
        y-=1
        rc.append((i%self.columns,i/self.columns))
      y/=2
      i+=1
    return rc
  def solve(self,board,verbose=True):
    self.problem=solver(self.sudoku_topology,self.output,verbose)
    self.read(board)
    try:
      self.problem.solve()
    except "Inconsistent":
      print "Board is inconsistent"
    if not verbose:
      self.output()
  def read(self,board):
    rc=[]
    distinct=0
    board=newline.sub("\n",board)
    for i in range(len(board)):
      x=i%(self.columns+1) # because of new-lines
      y=i/(self.columns+1)
      c=board[i]
      if x<self.columns and c!=' ':
        if c not in self.dictionary:
          if distinct==self.numbers:
            print "Here are some of the elements appearing on the board:"
            print self.dictionary.keys()
            raise "Board contains too many distinct elements."
          val=distinct
          self.dictionary[c]=val
          distinct+=1
        else:
          val=self.dictionary[c]
        vec=[int(j==val) for j in range(self.numbers)]
        rc.append((self.pos_to_num(x,y),vec[:],vec[:]))
    self.problem.oracle=rc
    self.inv_dict=range(self.numbers)
    for i in self.dictionary:
      self.inv_dict[self.dictionary[i]]=i
  def locnames(self,locs):
    rc=[]
    for i in locs:
      x=i[0]
      y=i[1]
      if self.n==3:
        x="ABCDEFGHI"[x]
        y="987654321"[y]
        rc.append(x+y)
      else:
        rc.append(str((x+1,self.columns-y)))
    rc.sort()
    if len(rc)>1:
      rc[-1:-1]=["and"]
    return " ".join(rc)
  def value_list_to_string(self,list):
    rc=[self.inv_dict[x] for x in list]
    rc.sort()
    if len(rc)>1:
      rc[-1:-1]=["and"]
    return " ".join(rc)
  def num_to_string(self,num):
    return self.locnames(self.num_to_list(num))
  def output(self):
    rc=""
    for y in range(self.rows):
      for x in range(self.columns):
        num=self.pos_to_num(x,y)
        mn=self.problem.constraints[num].mn
        if sum(mn)==1:
          rc+=self.inv_dict[mn.index(1)]
        else:
          rc+=' '
      rc+='\n'
    print rc
  def explain(self,position_string_list,value_string):
    value=self.dictionary[value_string]
    if type(position_string_list)==type(""):
      position_string_list=position_string_list.split(" ")
    elif type(position_string_list)!=type([]):
      position_string_list=[position_string_list]
    position_list=[]
    for position_string in position_string_list:
      if type(position_string)==type(""):
        x=ord(position_string[0])-ord('a')
        if (x<0) or (x>=self.numbers) or position_string[0]>'z':
          x=ord(position_string[0])-ord('A')
          if (x<0) or (x>=self.numbers) or position_string[0]>'Z':
            raise "Unknown position format. Use \"D6 E5\" or [(4,6),(5,5)] type format."
        y=self.numbers-(ord(position_string[1])-ord('0'))
        if (y<0) or (y>=self.numbers):
          raise "Unknown position format. Use \"D6 E5\" or [(4,6),(5,5)] type format."
        position_list.append((x,y))
      else:
        x=position_string[0]-1
        y=self.numbers-position_string[1]
        position_list.append((x,y))
    position=self.list_to_num(position_list)
    explanation=self.problem.constraints[position].explain
    for_val=[x for x in explanation if x[1]==value]
    if len(for_val)==0:
      print "It is as yet undetermined whether this list of positions contains",value_string,"or not."
    else:
      for_val=for_val[0]
      if for_val[0]==0:
        print "The position list does not contain",value_string,"."
        if for_val[2]==0:
          print "This was discovered by elimination, after finding that it must contain "+self.value_list_to_string([x[1] for x in explanation if x[0]==1])+"."
      else:
        print "The position list contains",value_string,"."
        if for_val[2]==0:
          print "This was discovered by elimination, after finding that it cannot contain "+self.value_list_to_string([x[1] for x in explanation if x[0]==0])+"."
      if for_val[2]==1:
        print "This is true because the list "+self.num_to_string(for_val[3])+" doesn't have it."
      elif for_val[2]==2:
        if for_val[0]==0:
          print "This is true because the list "+self.num_to_string(for_val[3])+" has it."
        else:
          print "This is true because the list "+self.num_to_string(position+for_val[3])+" has it, but the list "+self.num_to_string(for_val[3])+" alone does not."
      elif for_val[2]==3:
        print "This is true because the list "+self.num_to_string(for_val[3])+" has it."
      elif for_val[2]==4:
        print "This was discovered while reading the board."

if __name__=="__main__":
  p=sudoku()
  p.solve(sys.stdin.read())

